import os
import pathlib
import subprocess
from pathlib import Path
from queue import Queue
from threading import Thread

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from uvicorn import Config, Server

from jina import __version__, __resources_path__
from jina.logging.logger import JinaLogger
from .excepts import (
    RequestValidationError,
    Runtime400Exception,
    daemon_runtime_exception_handler,
    validation_exception_handler,
)
from .parser import get_main_parser, _get_run_args

jinad_args = get_main_parser().parse_args([])
daemon_logger = JinaLogger('DAEMON', **vars(jinad_args))

__task_queue__ = Queue()
__root_workspace__ = jinad_args.workspace
__rootdir__ = str(Path(__file__).parent.parent.absolute())
__dockerfiles__ = str(Path(__file__).parent.absolute() / 'Dockerfiles')


def _get_app(mode=None):
    from .api.endpoints import router

    app = FastAPI(
        title='JinaD (Daemon)',
        description='REST interface for managing distributed Jina',
        version=__version__,
        openapi_tags=[
            {
                'name': 'daemon',
                'description': 'API to manage the Daemon',
            },
        ],
    )
    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )

    app.include_router(router)
    app.add_exception_handler(Runtime400Exception, daemon_runtime_exception_handler)
    app.add_exception_handler(RequestValidationError, validation_exception_handler)

    if mode is None:
        from .api.endpoints import flows, pods, peas, logs, workspaces

        app.include_router(logs.router)
        app.include_router(peas.router)
        app.include_router(pods.router)
        app.include_router(flows.router)
        app.include_router(workspaces.router)
        app.openapi_tags.extend(
            [
                {
                    'name': 'flows',
                    'description': 'API to manage Flows',
                },
                {
                    'name': 'pods',
                    'description': 'API to manage Pods',
                },
                {
                    'name': 'peas',
                    'description': 'API to manage Peas',
                },
                {
                    'name': 'logs',
                    'description': 'API to stream Logs',
                },
                {
                    'name': 'workspaces',
                    'description': 'API to manage Workspaces',
                },
            ]
        )
    elif mode == 'pod':
        from .api.endpoints.partial import pod

        app.include_router(pod.router)
        app.openapi_tags.append(
            {
                'name': 'pod',
                'description': 'API to manage a Pod',
            }
        )
    elif mode == 'pea':
        from .api.endpoints.partial import pea

        app.include_router(pea.router)
        app.openapi_tags.append(
            {
                'name': 'pea',
                'description': 'API to manage a Pea',
            },
        )
    elif mode == 'flow':
        from .api.endpoints.partial import flow

        app.include_router(flow.router)
        app.openapi_tags.append(
            {
                'name': 'flow',
                'description': 'API to manage a Flow',
            }
        )

    return app


def _start_uvicorn(app: 'FastAPI'):
    config = Config(
        app=app,
        host=jinad_args.host,
        port=jinad_args.port_expose,
        loop='uvloop',
        log_level='error',
    )
    server = Server(config=config)
    server.run()

    from jina import __stop_msg__

    daemon_logger.success(__stop_msg__)


def _start_fluentd():
    daemon_logger.info('starting fluentd...')
    cfg = os.path.join(__resources_path__, 'fluent.conf')
    try:
        fluentd_proc = subprocess.Popen(
            ['fluentd', '-c', cfg],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            bufsize=0,
            universal_newlines=True,
        )
        for line in fluentd_proc.stdout:
            daemon_logger.debug(f'fluentd: {line.strip()}')
    except FileNotFoundError:
        daemon_logger.warning('fluentd not found locally, jinad cannot stream logs!')
        jinad_args.no_fluentd = True


def _start_consumer():
    from .tasks import ConsumerThread

    ConsumerThread().start()


def _update_default_args():
    global jinad_args, __root_workspace__
    jinad_args = _get_run_args()
    __root_workspace__ = '/workspace' if jinad_args.mode else jinad_args.workspace


def main():
    """Entrypoint for jinad"""
    _update_default_args()
    pathlib.Path(__root_workspace__).mkdir(parents=True, exist_ok=True)
    if not jinad_args.no_fluentd:
        Thread(target=_start_fluentd, daemon=True).start()
    _start_consumer()
    _start_uvicorn(app=_get_app(mode=jinad_args.mode))



"""Argparser module for hub push"""
from ..helper import add_arg_group


def mixin_hub_pull_parser(parser):
    """Add the arguments for hub pull to the parser
    :param parser: the parser configure
    """

    def hub_uri(uri: str) -> str:
        from ...hubble.helper import parse_hub_uri

        parse_hub_uri(uri)
        return uri

    gp = add_arg_group(parser, title='Pull')
    gp.add_argument(
        'uri',
        type=hub_uri,
        help='The URI of the executor to pull (e.g., jinahub[+docker]://UUID8)',
    )
    gp.add_argument(
        '--install-requirements',
        action='store_true',
        default=False,
        help='If set, install `requirements.txt` in the Hub Executor bundle to the local system',
    )



from enum import Enum
from typing import Dict

from pydantic import BaseModel

from .containers import ContainerItem, ContainerStoreStatus
from .custom import build_pydantic_model
from .id import DaemonID
from .workspaces import WorkspaceItem, WorkspaceStoreStatus

FlowModel = build_pydantic_model(model_name='FlowModel', module='flow')
PodModel = build_pydantic_model(model_name='PodModel', module='pod')
PeaModel = build_pydantic_model(model_name='PeaModel', module='pea')


class DaemonStatus(BaseModel):
    """Pydantic model for DaemonStatus"""

    jina: Dict
    envs: Dict
    workspaces: WorkspaceStoreStatus
    peas: ContainerStoreStatus
    pods: ContainerStoreStatus
    flows: ContainerStoreStatus
    used_memory: str


class UpdateOperationEnum(Enum):
    """Represents the type of operation to perform in the update

    We consider these an `update` operation since they **change** the underlying state
    """

    rolling_update = 'rolling_update'
    dump = 'dump'



import requests

from .containers import ContainerStore
from ..excepts import Runtime400Exception


class PeaStore(ContainerStore):
    """A Store of Peas spawned by as Containers by Daemon"""

    _kind = 'pea'

    def _add(self):
        try:
            self._logger.debug(
                f'sending POST request to mini-jinad on {self.host}/{self._kind}'
            )
            r = requests.post(url=f'{self.host}/{self._kind}', json=self.params)
            if r.status_code != requests.codes.created:
                raise Runtime400Exception(
                    f'{self._kind.title()} creation failed: {r.json()}'
                )
            return r.json()
        except requests.exceptions.RequestException as ex:
            self._logger.error(f'{ex!r}')
            raise Runtime400Exception(
                f'{self._kind.title()} deletion failed. request timed out'
            )

    def _update(self):
        # TODO
        pass

    def _delete(self, host, **kwargs):
        """Sends a delete request to terminate the Pea & remove the container

        :param host: host of mini-jinad
        :param kwargs: keyword args
        :raises Runtime400Exception: if deletion fails
        """
        try:
            self._logger.debug(
                f'sending DELETE request to mini-jinad on {host}/{self._kind}'
            )
            r = requests.delete(url=f'{host}/{self._kind}')
            if r.status_code != requests.codes.ok:
                raise Runtime400Exception(
                    f'{self._kind.title()} deletion failed: {r.json()}'
                )
        except requests.exceptions.RequestException as ex:
            raise Runtime400Exception(
                f'{self._kind.title()} deletion failed. request timed out'
            )
